#!/usr/bin/env python
import os
import os.path
import sys
import django
from glob import glob
from django.conf import settings
from django.test.utils import get_runner

def cleanMedia():
    mediaDir = settings.MEDIA_ROOT
    imgDir = os.path.join(mediaDir, "images")
    origImgDir = os.path.join(mediaDir, "original_images")
    for path in glob(os.path.join(imgDir, "*")) + glob(os.path.join(origImgDir, "*")):
        try:
            os.remove(path)
        except:
            print("Error deleting", path)

def runPytest():
    import pytest
    opts = []
    parallelOpt = ""
    labels = []
    for arg in sys.argv[1:]:
        if arg.startswith("-"):
            opts.append(arg)
            if arg.startswith("-n"):
                parallelOpt = arg
        else:
            # convert unittest "test_ical.TestImport.testMeetup" format to
            # "ls/joyous/test_ical.py::TestImport::testMeetup"
            path = arg.split(".")
            if len(path) >= 2 and path[1] == "py":
                del path[1]
            path[0] += ".py"
            labels.append("ls/joyous/tests/"+"::".join(path))
    if not parallelOpt:
        # default to running in parallel
        opts.append("-n2")
    errCode = pytest.main(opts + labels)
    return errCode

def runDjangoTest():
    cleanMedia()
    verbosity = 1
    if "-v" in sys.argv or "--verbose" in sys.argv:
        verbosity = 2
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner(top_level="ls/joyous",
                             verbosity=verbosity,
                             keepdb=False)
    labels = ["ls.joyous.tests."+arg for arg in sys.argv[1:]
              if not arg.startswith("-")]
    if not labels:
        labels = ["ls.joyous.tests"]
    failures = test_runner.run_tests(labels)
    return failures

def coverage():
    from coverage import Coverage
    cover = Coverage()
    cover.start()
    failures = runDjangoTest()
    cover.stop()
    cover.save()
    cover.html_report()
    return failures

def main():
    doRunCoverage = "--coverage" in sys.argv
    os.environ['DJANGO_SETTINGS_MODULE'] = 'ls.joyous.tests.settings'
    if "--pytest" in sys.argv:
        sys.argv.remove("--pytest")
        if doRunCoverage:
            sys.argv.remove("--coverage")
            sys.argv.append("--cov=ls/joyous")
            sys.argv.append("--cov-report=html")
        failures = runPytest()
    else:
        if doRunCoverage:
            failures = coverage()
        else:
            failures = runDjangoTest()
    return failures

if __name__ == "__main__":
    failures = main()
    sys.exit(failures)



# ------------------------------------------------------------------------------
# Test google Format
# see also test_gevents.py
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from io import BytesIO
from icalendar import vDatetime
from django.contrib.auth.models import User
from django.contrib.messages.storage.fallback import FallbackStorage
from django.test import TestCase, RequestFactory
from django.utils import timezone
from wagtail.core.models import Site, Page
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, CancellationPage, MultidayRecurringEventPage,
        ExtraInfoPage)
from ls.joyous.models import getAllEvents
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import WEEKLY, MONTHLY, MO, TU, WE, SA, FR
from ls.joyous.formats.google import GoogleCalendarHandler
from freezegun import freeze_time
from .testutils import datetimetz

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        Site.objects.update(hostname="joy.test")
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3(R3t')
        self.requestFactory = RequestFactory()
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.handler = GoogleCalendarHandler()

    def _getRequest(self, path="/"):
        request = self.requestFactory.get(path)
        request.user = self.user
        request.site = self.home.get_site()
        request.session = {}
        request._messages = FallbackStorage(request)
        request.POST = request.POST.copy()
        request.POST['action-publish'] = "action-publish"
        return request

    def testServeSimple(self):
        page = SimpleEventPage(owner = self.user,
                               slug  = "baseball-game",
                               title = "Baseball Game",
                               date  = dt.date(2014,7,30),
                               time_from = dt.time(13),
                               time_to   = dt.time(16),
                               tz = pytz.timezone("America/Los_Angeles"))
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        response = self.handler.serve(page,
                                      self._getRequest("/events/baseball-game/"
                                                       "?format=google"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url,
                         "http://www.google.com/calendar/event?action=TEMPLATE&"
                         "text=Baseball+Game&dates=20140730T130000%2F20140730T160000&"
                         "ctz=America%2FLos_Angeles")

    def testServeMultiday(self):
        page = MultidayEventPage(owner = self.user,
                                 slug  = "food-festival",
                                 title = "Food Festival",
                                 date_from = dt.date(2018,11,2),
                                 date_to   = dt.date(2018,11,5),
                                 tz = pytz.timezone("Pacific/Niue"))
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        response = self.handler.serve(page,
                                      self._getRequest("/events/food-festival/"
                                                       "?format=google"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url,
                         "http://www.google.com/calendar/event?action=TEMPLATE&"
                         "text=Food+Festival&dates=20181102T000000%2F20181105T235959&"
                         "ctz=Pacific%2FNiue")

    def testServeRecurring(self):
        page = RecurringEventPage(owner   = self.user,
                                  slug    = "lunch-and-code",
                                  title   = "Lunch and Code",
                                  repeat    = Recurrence(dtstart=dt.date(2017,3,15),
                                                         freq=MONTHLY,
                                                         byweekday=[FR(-1)]),
                                  time_from = dt.time(12),
                                  time_to   = dt.time(13),
                                  tz = pytz.timezone("Pacific/Auckland"),
                                  location  = "70 Molesworth Street, Wellington")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        response = self.handler.serve(page,
                                      self._getRequest("/events/lunch-and-code/"
                                                       "?format=google"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url,
                         "http://www.google.com/calendar/event?action=TEMPLATE&"
                         "text=Lunch+and+Code&"
                         "location=70+Molesworth+Street%2C+Wellington&"
                         "dates=20170331T120000%2F20170331T130000&"
                         "ctz=Pacific%2FAuckland&"
                         "recur=RRULE%3AFREQ%3DMONTHLY%3BWKST%3DSU%3BBYDAY%3D-1FR")

    def testServeException(self):
        event = RecurringEventPage(slug      = "test-meeting",
                                   title     = "Test Meeting",
                                   repeat    = Recurrence(dtstart=dt.date(1988,1,1),
                                                          freq=WEEKLY,
                                                          byweekday=[MO,WE,FR]),
                                   time_from = dt.time(13),
                                   time_to   = dt.time(15,30))
        self.calendar.add_child(instance=event)
        event.save_revision().publish()
        info = ExtraInfoPage(owner = self.user,
                             overrides = event,
                             except_date = dt.date(1988,11,11),
                             extra_title = "System Demo",
                             extra_information = "<h3>System Demo</h3>")
        event.add_child(instance=info)
        info.save_revision().publish()
        response = self.handler.serve(info,
                                      self._getRequest("/events/test-meeting/1998-11-11/extra-info/"
                                                       "?format=google"))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url,
                         "http://www.google.com/calendar/event?action=TEMPLATE&"
                         "text=Test+Meeting&"
                         "dates=19880101T130000%2F19880101T153000&"
                         "ctz=Asia%2FTokyo&"
                         "recur=RRULE%3AFREQ%3DWEEKLY%3BWKST%3DSU%3BBYDAY%3DMO%2CWE%2CFR")

    def testServeUnsupported(self):
        response = self.handler.serve(self.home, self._getRequest("/"))
        self.assertIsNone(response)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous calendar models
# ------------------------------------------------------------------------------
import datetime as dt
import calendar
from django.conf import settings
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.http import Http404
from django import forms
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.core.models import Page
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import HelpPanel, FieldPanel, MultiFieldPanel
from wagtail.contrib.routable_page.models import RoutablePageMixin, route
from wagtail.search import index
from .. import __version__
from ..edit_handlers import ConcealedPanel
from ..holidays import Holidays
from ..utils.names import WEEKDAY_NAMES, MONTH_NAMES, MONTH_ABBRS
from ..utils.weeks import week_info, gregorian_to_week_date, num_weeks_in_year
from ..utils.weeks import weekday_abbr, weekday_name
from ..utils.mixins import ProxyPageMixin
from ..fields import MultipleSelectField
from . import (getAllEventsByDay, getAllEventsByWeek, getAllUpcomingEvents,
               getAllPastEvents, getEventFromUid, getAllEvents)

# ------------------------------------------------------------------------------
class CalendarPageForm(WagtailAdminPageForm):
    importHandler = None
    exportHandler = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def registerImportHandler(cls, handler):
        class Panel(ConcealedPanel):
            def _show(self):
                page = getattr(self, 'instance', None)
                if not page:
                    return False
                hasReq = hasattr(page, '__joyous_edit_request')
                if not hasReq:
                    return False
                # only a user with edit and publishing rights should be able
                # to import iCalendar files
                perms = page.permissions_for_user(self.request.user)
                return perms.can_publish() and perms.can_edit()

        # TODO support multiple formats?
        cls.importHandler = handler
        uploadWidget = forms.FileInput(attrs={'accept': "text/calendar"})
        cls.declared_fields['upload'] = forms.FileField(
                                            label=_("Upload"),
                                            required=False,
                                            widget=uploadWidget)
        cls.declared_fields['utc2local'] = forms.BooleanField(
                                            label=_("Convert UTC to localtime?"),
                                            required=False,
                                            initial=True)
        CalendarPage.settings_panels.append(Panel([
              FieldPanel('upload'),
              FieldPanel('utc2local'),
            ], heading=_("Import")))

    @classmethod
    def registerExportHandler(cls, handler):
        class Panel(ConcealedPanel):
            def _show(self):
                page = getattr(self, 'instance', None)
                return page and page.url is not None and page.live

        cls.exportHandler = handler
        CalendarPage.settings_panels.append(Panel([
              # TODO: annex the HelpPanel into ExportPanel?
              HelpPanel(template="joyous/edit_handlers/export_panel.html")
            ], heading=_("Export")))

    def save(self, commit=True):
        page = super().save(commit=False)
        request = getattr(page, '__joyous_edit_request', None)

        if self.importHandler and request:
            delattr(page, '__joyous_edit_request')
            utc2local = self.cleaned_data.get('utc2local')
            upload = self.cleaned_data.get('upload')
            if upload is not None:
                self.importHandler.load(page, request, upload, utc2local=utc2local)

        if commit:
            page.save()
        return page

# ------------------------------------------------------------------------------
DatePictures = {"YYYY":  r"((?:19|20)\d\d)",
                "MM":    r"(1[012]|0?[1-9])",
                "Mon":   r"({})".format("|".join(m.lower()[:3] for m in MONTH_ABBRS[1:])),
                "DD":    r"(3[01]|[12]\d|0?[1-9])",
                "WW":    r"(5[0-3]|[1-4]\d|0?[1-9])"}

EVENTS_VIEW_CHOICES = [('L', _("List View")),
                       ('W', _("Weekly View")),
                       ('M', _("Monthly View"))]

# ------------------------------------------------------------------------------
class CalendarPage(RoutablePageMixin, Page):
    """CalendarPage displays all the events which are in the same site."""
    class Meta:
        verbose_name = _("calendar page")
        verbose_name_plural = _("calendar pages")

    EventsPerPage = getattr(settings, "JOYOUS_EVENTS_PER_PAGE", 25)
    holidays = Holidays()
    subpage_types = ['joyous.SimpleEventPage',
                     'joyous.MultidayEventPage',
                     'joyous.RecurringEventPage',
                     'joyous.MultidayRecurringEventPage']
    base_form_class = CalendarPageForm

    intro = RichTextField(_("intro"), blank=True,
                          help_text=_("Introductory text."))
    view_choices = MultipleSelectField(_("view choices"), blank=True,
                                       default=["L","W","M"],
                                       choices=EVENTS_VIEW_CHOICES)
    default_view = models.CharField(_("default view"),
                                    default="M", max_length=15,
                                    choices=EVENTS_VIEW_CHOICES)

    search_fields = Page.search_fields[:]
    content_panels = Page.content_panels + [
        FieldPanel('intro', classname="full"),
        ]
    settings_panels = Page.settings_panels + [
        MultiFieldPanel([
            FieldPanel('view_choices'),
            FieldPanel('default_view')],
            heading=_("View Options")),
        ]

    @route(r"^$")
    @route(r"^{YYYY}/$".format(**DatePictures))
    def routeDefault(self, request, year=None):
        """Route a request to the default calendar view."""
        eventsView = request.GET.get('view', self.default_view)
        if eventsView in ("L", "list"):
            return self.serveUpcoming(request)
        elif eventsView in ("W", "weekly"):
            return self.serveWeek(request, year)
        else:
            return self.serveMonth(request, year)

    @route(r"^{YYYY}/{Mon}/$(?i)".format(**DatePictures))
    def routeByMonthAbbr(self, request, year, monthAbbr):
        """Route a request with a month abbreviation to the monthly view."""
        month = (DatePictures['Mon'].index(monthAbbr.lower()) // 4) + 1
        return self.serveMonth(request, year, month)

    @route(r"^month/$")
    @route(r"^{YYYY}/{MM}/$".format(**DatePictures))
    def serveMonth(self, request, year=None, month=None):
        """Monthly calendar view."""
        myurl = self.get_url(request)
        def myUrl(urlYear, urlMonth):
            if 1900 <= urlYear <= 2099:
                return myurl + self.reverse_subpage('serveMonth',
                                                    args=[urlYear, urlMonth])
        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        year = int(year)
        month = int(month)
        lastDay = dt.date(year, month, calendar.monthrange(year, month)[1])

        if year == today.year and month == today.month:
            weekYear, weekNum, dow = gregorian_to_week_date(today)
        else:
            weekYear, weekNum, dow = gregorian_to_week_date(dt.date(year, month, 7))
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[weekYear, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')

        prevMonth = month - 1
        prevMonthYear = year
        if prevMonth == 0:
            prevMonth = 12
            prevMonthYear -= 1

        nextMonth = month + 1
        nextMonthYear = year
        if nextMonth == 13:
            nextMonth = 1
            nextMonthYear += 1

        cxt = self._getCommonContext(request)
        cxt.update({'year':         year,
                    'month':        month,
                    'yesterday':    today - dt.timedelta(1),
                    'lastweek':     today - dt.timedelta(7),
                    'lastDay':      lastDay,
                    'prevMonthUrl': myUrl(prevMonthYear, prevMonth),
                    'nextMonthUrl': myUrl(nextMonthYear, nextMonth),
                    'prevYearUrl':  myUrl(year - 1, month),
                    'nextYearUrl':  myUrl(year + 1, month),
                    'weeklyUrl':    weeklyUrl,
                    'listUrl':      listUrl,
                    'thisMonthUrl': myUrl(today.year, today.month),
                    'monthName':    MONTH_NAMES[month],
                    'weekdayAbbr':  weekday_abbr,
                    'events':       self._getEventsByWeek(request, year, month)})
        cxt.update(self._getExtraContext("month"))
        return TemplateResponse(request,
                                "joyous/calendar_month.html",
                                cxt)

    @route(r"^week/$")
    @route(r"^{YYYY}/W{WW}/$".format(**DatePictures))
    def serveWeek(self, request, year=None, week=None):
        """Weekly calendar view."""
        myurl = self.get_url(request)
        def myUrl(urlYear, urlWeek):
            if (urlYear < 1900 or
                urlYear > 2099 or
                urlYear == 2099 and urlWeek == 53):
                return None
            if urlWeek == 53 and num_weeks_in_year(urlYear) == 52:
                urlWeek = 52
            return myurl + self.reverse_subpage('serveWeek',
                                                args=[urlYear, urlWeek])
        today = timezone.localdate()
        thisYear, thisWeekNum, dow = gregorian_to_week_date(today)
        if year is None: year = thisYear
        if week is None: week = thisWeekNum
        year = int(year)
        week = int(week)

        firstDay, lastDay, prevYearNumWeeks, yearNumWeeks = week_info(year, week)
        if week == 53 and yearNumWeeks == 52:
            raise Http404("Only 52 weeks in {}".format(year))

        eventsInWeek = self._getEventsByDay(request, firstDay, lastDay)
        if firstDay.year >= 1900:
            monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                      args=[firstDay.year, firstDay.month])
        else:
            monthlyUrl = myurl + self.reverse_subpage('serveMonth', args=[1900, 1])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')

        prevWeek = week - 1
        prevWeekYear = year
        if prevWeek == 0:
            prevWeek = prevYearNumWeeks
            prevWeekYear -= 1

        nextWeek = week + 1
        nextWeekYear = year
        if nextWeek > yearNumWeeks:
            nextWeek = 1
            nextWeekYear += 1

        cxt = self._getCommonContext(request)
        cxt.update({'year':         year,
                    'week':         week,
                    'yesterday':    today - dt.timedelta(1),
                    'prevWeekUrl':  myUrl(prevWeekYear, prevWeek),
                    'nextWeekUrl':  myUrl(nextWeekYear, nextWeek),
                    'prevYearUrl':  myUrl(year - 1, week),
                    'nextYearUrl':  myUrl(year + 1, week),
                    'thisWeekUrl':  myUrl(thisYear, thisWeekNum),
                    'monthlyUrl':   monthlyUrl,
                    'listUrl':      listUrl,
                    'weekName':     _("Week {weekNum}").format(weekNum=week),
                    'weekdayAbbr':  weekday_abbr,
                    'events':       [eventsInWeek]})
        cxt.update(self._getExtraContext("week"))
        return TemplateResponse(request,
                                "joyous/calendar_week.html",
                                cxt)

    @route(r"^day/$")
    @route(r"^{YYYY}/{MM}/{DD}/$".format(**DatePictures))
    def serveDay(self, request, year=None, month=None, dom=None):
        """The events of the day list view."""
        myurl = self.get_url(request)
        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        if dom is None: dom = today.day
        year = int(year)
        month = int(month)
        dom = int(dom)
        day = dt.date(year, month, dom)

        daysEvents = self._getEventsOnDay(request, day).all_events
        if len(daysEvents) == 1:
            event = daysEvents[0].page
            return redirect(event.get_url(request))
        eventsPage = self._paginate(request, daysEvents)

        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[year, month])
        weekYear, weekNum, dow = gregorian_to_week_date(dt.date(year, month, 7))
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[weekYear, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')

        cxt = self._getCommonContext(request)
        cxt.update({'year':         year,
                    'month':        month,
                    'dom':          dom,
                    'day':          day,
                    'monthlyUrl':   monthlyUrl,
                    'weeklyUrl':    weeklyUrl,
                    'listUrl':      listUrl,
                    'monthName':    MONTH_NAMES[month],
                    'weekdayName':  WEEKDAY_NAMES[day.weekday()],
                    'events':       eventsPage})
        cxt.update(self._getExtraContext("day"))
        return TemplateResponse(request,
                                "joyous/calendar_list_day.html",
                                cxt)

    @route(r"^upcoming/$")
    def serveUpcoming(self, request):
        """Upcoming events list view."""
        myurl = self.get_url(request)
        today = timezone.localdate()
        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[today.year, today.month])
        weekYear, weekNum, dow = gregorian_to_week_date(today)
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[weekYear, weekNum])
        listUrl = myurl + self.reverse_subpage('servePast')
        upcomingEvents = self._getUpcomingEvents(request)
        eventsPage = self._paginate(request, upcomingEvents)

        cxt = self._getCommonContext(request)
        cxt.update({'weeklyUrl':    weeklyUrl,
                    'monthlyUrl':   monthlyUrl,
                    'listUrl':      listUrl,
                    'events':       eventsPage})
        cxt.update(self._getExtraContext("upcoming"))
        return TemplateResponse(request,
                                "joyous/calendar_list_upcoming.html",
                                cxt)

    @route(r"^past/$")
    def servePast(self, request):
        """Past events list view."""
        myurl = self.get_url(request)
        today = timezone.localdate()
        monthlyUrl = myurl + self.reverse_subpage('serveMonth',
                                                  args=[today.year, today.month])
        weekYear, weekNum, dow = gregorian_to_week_date(today)
        weeklyUrl = myurl + self.reverse_subpage('serveWeek',
                                                 args=[weekYear, weekNum])
        listUrl = myurl + self.reverse_subpage('serveUpcoming')
        pastEvents = self._getPastEvents(request)
        eventsPage = self._paginate(request, pastEvents)

        cxt = self._getCommonContext(request)
        cxt.update({'weeklyUrl':    weeklyUrl,
                    'monthlyUrl':   monthlyUrl,
                    'listUrl':      listUrl,
                    'events':       eventsPage})
        cxt.update(self._getExtraContext("past"))
        return TemplateResponse(request,
                                "joyous/calendar_list_past.html",
                                cxt)

    @route(r"^mini/{YYYY}/{MM}/$".format(**DatePictures))
    def serveMiniMonth(self, request, year=None, month=None):
        """Serve data for the MiniMonth template tag."""
        if not request.is_ajax():
            raise Http404("/mini/ is for ajax requests only")

        today = timezone.localdate()
        if year is None: year = today.year
        if month is None: month = today.month
        year = int(year)
        month = int(month)

        cxt = self._getCommonContext(request)
        cxt.update({'year':         year,
                    'month':        month,
                    'calendarUrl':  self.get_url(request),
                    'monthName':    MONTH_NAMES[month],
                    'weekdayInfo':  zip(weekday_abbr, weekday_name),
                    'events':       self._getEventsByWeek(request, year, month)})
        cxt.update(self._getExtraContext("mini"))
        return TemplateResponse(request,
                                "joyous/includes/minicalendar.html",
                                cxt)

    @classmethod
    def can_create_at(cls, parent):
        return super().can_create_at(parent) and cls._allowAnotherAt(parent)

    @classmethod
    def _allowAnotherAt(cls, parent):
        """You can only create one of these pages per site."""
        site = parent.get_site()
        if site is None:
            return False
        return not cls.peers().descendant_of(site.root_page).exists()

    @classmethod
    def peers(cls):
        """Return others of the same concrete type."""
        contentType = ContentType.objects.get_for_model(cls)
        return cls.objects.filter(content_type=contentType)

    def _getCommonContext(self, request):
        cxt = self.get_context(request)
        cxt.update({'version':  __version__,
                    'themeCSS': getattr(settings, "JOYOUS_THEME_CSS", ""),
                    'today':    timezone.localdate()})
        return cxt

    def _getExtraContext(self, route):
        return {}

    def _getEventsOnDay(self, request, day):
        """Return all the events in this site for a given day."""
        return self._getEventsByDay(request, day, day)[0]

    def _getEventsByDay(self, request, firstDay, lastDay):
        """
        Return the events in this site for the dates given, grouped by day.
        """
        home = request.site.root_page
        return getAllEventsByDay(request, firstDay, lastDay,
                                 home=home, holidays=self.holidays)

    def _getEventsByWeek(self, request, year, month):
        """
        Return the events in this site for the given month grouped by week.
        """
        home = request.site.root_page
        return getAllEventsByWeek(request, year, month,
                                  home=home, holidays=self.holidays)

    def _getUpcomingEvents(self, request):
        """Return the upcoming events in this site."""
        home = request.site.root_page
        return getAllUpcomingEvents(request, home=home)

    def _getPastEvents(self, request):
        """Return the past events in this site."""
        home = request.site.root_page
        return getAllPastEvents(request, home=home)

    def _getEventFromUid(self, request, uid):
        """Try and find an event with the given UID in this site."""
        event = getEventFromUid(request, uid) # might raise exception
        home = request.site.root_page
        if event.get_ancestors().filter(id=home.id).exists():
            # only return event if it is in the same site
            return event

    def _getAllEvents(self, request):
        """Return all the events in this site."""
        home = request.site.root_page
        return getAllEvents(request, home=home)

    def _paginate(self, request, events):
        paginator = Paginator(events, self.EventsPerPage)
        try:
            eventsPage = paginator.page(request.GET.get('page'))
        except PageNotAnInteger:
            eventsPage = paginator.page(1)
        except EmptyPage:
            eventsPage = paginator.page(paginator.num_pages)
        return eventsPage

# ------------------------------------------------------------------------------
class SpecificCalendarPage(ProxyPageMixin, CalendarPage):
    """
    SpecificCalendarPage displays only the events which are its children
    """
    class Meta(ProxyPageMixin.Meta):
        verbose_name = _("specific calendar page")
        verbose_name_plural = _("specific calendar pages")

    is_creatable  = False  # creation is disabled by default

    @classmethod
    def _allowAnotherAt(cls, parent):
        """Don't limit creation."""
        return True

    def _getEventsByDay(self, request, firstDay, lastDay):
        """
        Return my child events for the dates given,  grouped by day.
        """
        return getAllEventsByDay(request, firstDay, lastDay,
                                 home=self, holidays=self.holidays)

    def _getEventsByWeek(self, request, year, month):
        """Return my child events for the given month grouped by week."""
        return getAllEventsByWeek(request, year, month,
                                  home=self, holidays=self.holidays)

    def _getUpcomingEvents(self, request):
        """Return my upcoming child events."""
        return getAllUpcomingEvents(request, home=self)

    def _getPastEvents(self, request):
        """Return my past child events."""
        return getAllPastEvents(request, home=self)

    def _getEventFromUid(self, request, uid):
        """Try and find a child event with the given UID."""
        event = getEventFromUid(request, uid) # might raise exception
        if event.get_ancestors().filter(id=self.id).exists():
            # only return event if it is a descendant
            return event

    def _getAllEvents(self, request):
        """Return all my child events."""
        return getAllEvents(request, home=self)

# ------------------------------------------------------------------------------
class GeneralCalendarPage(ProxyPageMixin, CalendarPage):
    """
    GeneralCalendarPage displays all the events no matter where they are
    """
    class Meta(ProxyPageMixin.Meta):
        verbose_name = _("general calendar page")
        verbose_name_plural = _("general calendar pages")

    is_creatable  = False  # creation is disabled by default

    @classmethod
    def _allowAnotherAt(cls, parent):
        """You can only create one of these pages."""
        return not cls.peers().exists()

    def _getEventsByDay(self, request, firstDay, lastDay):
        """
        Return all events for the dates given, grouped by day.
        """
        return getAllEventsByDay(request, firstDay, lastDay,
                                 holidays=self.holidays)

    def _getEventsByWeek(self, request, year, month):
        """Return all events for the given month grouped by week."""
        return getAllEventsByWeek(request, year, month, holidays=self.holidays)

    def _getUpcomingEvents(self, request):
        """Return all the upcoming events."""
        return getAllUpcomingEvents(request)

    def _getPastEvents(self, request):
        """Return all the past events."""
        return getAllPastEvents(request)

    def _getEventFromUid(self, request, uid):
        """Try and find an event with the given UID."""
        return getEventFromUid(request, uid) # might raise exception

    def _getAllEvents(self, request):
        """Return all the events."""
        return getAllEvents(request)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Formatting weeks
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.test import TestCase
import ls.joyous.utils.weeks
from ls.joyous.utils.weeks import (
        _iso_year_start, _iso_info, _iso_num_weeks,
        _gregorian_to_iso, _iso_to_gregorian, _iso_week_of_month)
from ls.joyous.utils.weeks import (
        _ssweek_year_start, _ssweek_info, _ssweek_num_weeks,
        _gregorian_to_ssweek, _ssweek_to_gregorian, _ssweek_of_month)
from django.utils.translation import override
from django.utils.formats import get_format
import importlib

# ------------------------------------------------------------------------------
class TestMondayStartingWeek(TestCase):
    def testYearStart(self):
        self.assertEqual(_iso_year_start(1991), dt.date(1990,12,31))
        self.assertEqual(_iso_year_start(2006), dt.date(2006,1,2))
        self.assertEqual(_iso_year_start(2020), dt.date(2019,12,30))
        self.assertEqual(_iso_year_start(2036), dt.date(2035,12,31))

    def testNumWeeksInYear(self):
        self.assertEqual(_iso_num_weeks(1991), 52)
        self.assertEqual(_iso_num_weeks(2006), 52)
        self.assertEqual(_iso_num_weeks(2020), 53)
        self.assertEqual(_iso_num_weeks(2036), 52)

    def testWeekOfMonth(self):
        self.assertEqual(_iso_week_of_month(dt.date(1990,5,31)),  4)
        self.assertEqual(_iso_week_of_month(dt.date(2005,1,2)),   0)
        self.assertEqual(_iso_week_of_month(dt.date(2021,12,13)), 2)
        self.assertEqual(_iso_week_of_month(dt.date(2030,10,8)),  1)

    def testGregorianToWeekDate(self):
        self.assertEqual(_gregorian_to_iso(dt.date(1991,5,25)),  (1991,21,6))
        self.assertEqual(_gregorian_to_iso(dt.date(2007,1,2)),   (2007,1,2))
        self.assertEqual(_gregorian_to_iso(dt.date(2022,12,13)), (2022,50,2))
        self.assertEqual(_gregorian_to_iso(dt.date(2035,12,31)), (2036,1,1))

    def testWeekToGregorianDate(self):
        self.assertEqual(_iso_to_gregorian(1990,2,6),  dt.date(1990,1,13))
        self.assertEqual(_iso_to_gregorian(2005,20,1), dt.date(2005,5,16))
        self.assertEqual(_iso_to_gregorian(2020,53,5), dt.date(2021,1,1))
        self.assertEqual(_iso_to_gregorian(2035,6,7),  dt.date(2035,2,11))

    def testWeekInfo(self):
        # (first_day, last_day, prev_year_num_weeks, year_num_weeks)
        self.assertEqual(_iso_info(1991,2),
                         (dt.date(1991,1,7),  dt.date(1991,1,13), 52, 52))
        self.assertEqual(_iso_info(2006,20),
                         (dt.date(2006,5,15), dt.date(2006,5,21), 52, 52))
        self.assertEqual(_iso_info(2020,40),
                         (dt.date(2020,9,28), dt.date(2020,10,4), 52, 53))
        self.assertEqual(_iso_info(2036,21),
                         (dt.date(2036,5,19), dt.date(2036,5,25), 52, 52))

# ------------------------------------------------------------------------------
class TestSundayStartingWeek(TestCase):
    def testYearStart(self):
        self.assertEqual(_ssweek_year_start(1991), dt.date(1990,12,30))
        self.assertEqual(_ssweek_year_start(2006), dt.date(2006,1,1))
        self.assertEqual(_ssweek_year_start(2020), dt.date(2019,12,29))
        self.assertEqual(_ssweek_year_start(2036), dt.date(2035,12,30))

    def testNumWeeksInYear(self):
        self.assertEqual(_ssweek_num_weeks(1991), 52)
        self.assertEqual(_ssweek_num_weeks(2006), 52)
        self.assertEqual(_ssweek_num_weeks(2020), 53)
        self.assertEqual(_ssweek_num_weeks(2036), 53)

    def testWeekOfMonth(self):
        self.assertEqual(_ssweek_of_month(dt.date(1990,5,31)),  4)
        self.assertEqual(_ssweek_of_month(dt.date(2005,1,2)),   1)
        self.assertEqual(_ssweek_of_month(dt.date(2021,12,13)), 2)
        self.assertEqual(_ssweek_of_month(dt.date(2030,10,8)),  1)

    def testGregorianToWeekDate(self):
        self.assertEqual(_gregorian_to_ssweek(dt.date(1991,5,25)),  (1991,21,6))
        self.assertEqual(_gregorian_to_ssweek(dt.date(2007,1,2)),   (2007,1,2))
        self.assertEqual(_gregorian_to_ssweek(dt.date(2022,12,13)), (2022,50,2))
        self.assertEqual(_gregorian_to_ssweek(dt.date(2035,12,31)), (2036,1,1))
        self.assertEqual(_gregorian_to_ssweek(dt.date(2019,12,1)),  (2019,49,7))
        self.assertEqual(_gregorian_to_ssweek(dt.date(2019,12,29)), (2020,1,7))

    def testWeekToGregorianDate(self):
        self.assertEqual(_ssweek_to_gregorian(1990,2,6),  dt.date(1990,1,12))
        self.assertEqual(_ssweek_to_gregorian(2005,20,1), dt.date(2005,5,15))
        self.assertEqual(_ssweek_to_gregorian(2020,53,5), dt.date(2020,12,31))
        self.assertEqual(_ssweek_to_gregorian(2035,6,7),  dt.date(2035,2,10))

    def testWeekInfo(self):
        # (first_day, last_day, prev_year_num_weeks, year_num_weeks)
        self.assertEqual(_ssweek_info(1991,2),
                         (dt.date(1991,1,6),  dt.date(1991,1,12), 52, 52))
        self.assertEqual(_ssweek_info(2006,20),
                         (dt.date(2006,5,14), dt.date(2006,5,20), 52, 52))
        self.assertEqual(_ssweek_info(2020,40),
                         (dt.date(2020,9,27), dt.date(2020,10,3), 52, 53))
        self.assertEqual(_ssweek_info(2036,21),
                         (dt.date(2036,5,18), dt.date(2036,5,24), 52, 53))

# ------------------------------------------------------------------------------
class TestSetting(TestCase):
    @override('en-gb')
    def testMondayStartingWeek(self):
        # FIRST_DAY_OF_WEEK is Monday for Great Britain
        self.assertEqual(get_format('FIRST_DAY_OF_WEEK'), 1)
        importlib.reload(ls.joyous.utils.weeks)
        from ls.joyous.utils.weeks import (week_info, num_weeks_in_year,
                gregorian_to_week_date, week_of_month, weekday_abbr, weekday_name,
                _iso_info, _iso_num_weeks, _gregorian_to_iso, _iso_week_of_month)
        self.assertIs(week_info, _iso_info)
        self.assertIs(num_weeks_in_year, _iso_num_weeks)
        self.assertIs(gregorian_to_week_date, _gregorian_to_iso)
        self.assertIs(week_of_month, _iso_week_of_month)
        self.assertEqual(weekday_abbr, ("Mon","Tue","Wed","Thu","Fri","Sat","Sun"))
        self.assertEqual(weekday_name, ("Monday","Tuesday","Wednesday","Thursday",
                                        "Friday","Saturday","Sunday"))

    @override('en-au')
    def testSundayStartingWeek(self):
        # FIRST_DAY_OF_WEEK is Sunday for Australia
        self.assertEqual(get_format('FIRST_DAY_OF_WEEK'), 0)
        importlib.reload(ls.joyous.utils.weeks)
        from ls.joyous.utils.weeks import (week_info, num_weeks_in_year,
                gregorian_to_week_date, week_of_month, weekday_abbr, weekday_name,
                _ssweek_info, _ssweek_num_weeks, _gregorian_to_ssweek, _ssweek_of_month)
        self.assertIs(week_info, _ssweek_info)
        self.assertIs(num_weeks_in_year, _ssweek_num_weeks)
        self.assertIs(gregorian_to_week_date, _gregorian_to_ssweek)
        self.assertIs(week_of_month, _ssweek_of_month)
        self.assertEqual(weekday_abbr, ("Sun","Mon","Tue","Wed","Thu","Fri","Sat"))
        self.assertEqual(weekday_name, ("Sunday","Monday","Tuesday","Wednesday","Thursday",
                                        "Friday","Saturday"))

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous Holidays
# ------------------------------------------------------------------------------
import datetime as dt
from collections import OrderedDict
from django.conf import settings
from .parser import parseHolidays

class Holidays:
    """Defines what holidays are celebrated on what dates."""
    def __init__(self):
        self.simple = {}
        self.srcs = [ self.simple ]
        self._parseSettings()

    def _parseSettings(self):
        holidaySettings = getattr(settings, "JOYOUS_HOLIDAYS", "")
        if holidaySettings:
            hols = parseHolidays(holidaySettings)
            if hols is not None:
                self.register(hols)

    def register(self, src):
        """Register a new source of holiday data."""
        self.srcs.append(src)

    def add(self, date, value):
        """Add a holiday to an individual date."""
        oldValue = self.simple.get(date)
        if oldValue:
            if oldValue not in value and value not in oldValue:
                self.simple[date] = "{}, {}".format(oldValue, value)
        else:
            self.simple[date] = value

    def get(self, date):
        """Get all the holidays that are celebrated on this date."""
        holidays = []
        for src in self.srcs:
            # get from python-holidays and other dict type srcs
            getHoliday = getattr(src, "get", None)
            if not getHoliday:
                # get from workalendar srcs
                getHoliday = getattr(src, "get_holiday_label")
            holiday = getHoliday(date)
            if holiday:
                holidays.extend(holiday.split(", "))
        holidays = list(OrderedDict.fromkeys(holidays))   # remove duplicates
        return ", ".join(holidays)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
